#!/usr/bin/python
# -*- coding: utf-8 -*-
"""IBK Simulation Run Support

Provides classes DelphinSolver.

"""

__author__ = "Andreas Nicolai"
__email__ = "andreas.nicolai@tu-dresden.de"
__copyright__ = "Copyright 2014, IBK, TU Dresden"
__license__ = "BSD-compatible (see LICENSE.txt)"
__version__ = "0.1.0"

from simulate import SimJob

class SolverData:
	"""Holds identification data for a single solver.
	Stored is the full version number, a descriptive text which is 
	used in the protocol and the full path to the solver executable.
	In future there may be an additional environment information, in case that the solver requires
	additional library search paths.

	**Example:**
	.. code-block:: python

	   >>> SolverData("5.6.9_rc1", "Delphin 5.6 Release candidate solver", "/usr/bin/IBK/DelphinSolver")
	"""
	def __init__(self, detailedVersion, description, path):
		self.detailedVersion = detailedVersion
		self.description = description
		self.path = path

	def __repr__(self):
		return "{}('{}','{}','{}')".format(self.__class__.__name__,
		                                   self.detailedVersion, self.description, self.path)

	def __str__(self):
		return "('{}','{}','{}')".format(self.detailedVersion, self.description, self.path)



def solvers():
	"""Returns a dictionary of solvers.
	Key of dictionary is solver ID (without version number).
	Value is a solver dictionary of all solvers known by this ID.

	**Example:**

	.. code-block:: python

	   >>> from IBK import Solvers

	   >>> print Solvers.solvers()
	   >>> 
	"""
	simulators = {}

	# TODO : read user configuration
	simulators["Delphin"] = {}
	simulators["Delphin"]["5.6"] = []
	simulators["Delphin"]["5.6"].append( SolverData("5.6.9","Last Delphin 5.6 Release", "/home/ghorwin/bin/ds56") )

	simulators["Therakles"] = {}
	simulators["Therakles"]["2.1"] = []
	simulators["Therakles"]["2.1"].append( SolverData("2.1.2", "Current Therakles Version", "/home/ghorwin/Therakles-2.1.2/bin/TheraklesSolver") )
	return simulators


def findSolver(solverID, majorVersion, detailedVersion):
	"""Searches for a matching solver in the list of registered solves and returns
	a SolverData object. If no matching solver is found, the function returns None and
	writes out an error message. If several matching solvers are found, the function
	returns the first and issues a warning message.

	"""
	solvs = solvers()
	if solverID in solvs:
		matching_solvers = solvs[solverID]
	else:
		print "Error: No solver with ID '{}' registered.".format(solverID)
		return None

	if majorVersion in matching_solvers:
		matching_solvers = matching_solvers[majorVersion]
	else:
		print "Error: No '{}' solver with major version '{}' registered.".format(solverID, majorVersion)
		return None

	# if no patch version is selected simply use the first matching solver
	if detailedVersion != None and len(detailedVersion)>0:
		matching_solvers = [s for s in matching_solvers if s.detailedVersion == detailedVersion]
	if len(matching_solvers) == 0:
		print "Error: No '{}' solver {} with detailed version '{}' registered.".format(solverID, majorVersion, detailedVersion)
		return None
	if len(matching_solvers) > 1:
		if detailedVersion != None and len(detailedVersion)>0:
			print "Warning: Multiple '{}' solvers {} with detailed version '{}' registered, selecting the first in list.".format(solverID, majorVersion, detailedVersion)
	return matching_solvers[0]


class DelphinSolver(SimJob):
	"""Convenience function to construct simulation objects that use a Delphin solver as simulator.
	**Example:**
	.. code-block:: python

	   >>> # create solver object
	   >>> solver = DelphinSolver("5.6","5.6.9_rc1", "/home/ghorwin/sims/cases/test1.dpj")

	   >>> # run solver
	   >>> solver.run()
	"""
	def __init__(self, majorVersion, detailedVersion, project):
		# retrieve selected solver info from available solvers
		solver = findSolver("Delphin", majorVersion, detailedVersion)
		if solver==None:
			SimJob.__init__(self, "")
			return

		# select default command line switches based on major solver version
		default_cmd_args = ["-x"]
		if majorVersion.find("6.") == 0:
			default_cmd_args.append("--verbosity-level=1")
		else:
			default_cmd_args.append("-v1")
		SimJob.__init__(self, [solver.path] + default_cmd_args + [project])
