#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
Annex60 Python Library - Simulation run support
-----------------------------------------------

Provides classes SimJob and SimRunner.

SimJob defines the functionality and interface for
a simulation job and associated pre- and post-processing steps.

SimRunner provides the functionality to run several simulation jobs in parallel.
"""

__author__ = "Andreas Nicolai"
__email__ = "andreas.nicolai@tu-dresden.de"
__copyright__ = "Copyright 2014, IBK, TU Dresden"
__license__ = "BSD-compatible (see LICENSE.txt)"
__version__ = "0.1.0"


import sys              # for Python version checking
import subprocess, os   # for subprocess and os.path
import platform         # for checking for Windows OS
import threading,Queue  # for parallel execution


# Check the Python version
major, minor1, minor2, s, tmp = sys.version_info
if not (major == 2 and minor1 == 7):
	raise ImportError('Currently, Python 2.7 is required.')

__all__ = ["SimJob", "SimRunner"]


class SimJob:
	"""
	This class stores information about executing a simulation job.
	Basically, this class stores the information to spawn a solver/simulator
	process via subprocess.Popen().

	**Example:**

	.. code-block:: python

	   >>> from simulate import *

	   >>> # create job
	   >>> job = SimJob(['~/bin/DelphinSolver','-x','--verbosity-level=2','~/cases/D6/test/wall1.d6p'])

	   >>> # execute job manually
	   >>> job.run()

	   >>> # result code of operation
	   >>> print job.return_code
	   0

	   >>> # in case of error, inquire error string
	   >>> if job.return_code != 0:
	   >>>     print job.error_string
	"""

	def __init__(self, cmdline, working_dir=None, environment=None):
		"""Constructs a new SimJob object.

		**Arguments:**

		- *cmdline*: Command line as array of arguments, see subprocess.Popen()

		- *working_dir*: None, or a valid path that should be used as working directory

		- *environment*: None, or a valid environment specification that should be used
		  by the spawned process
		"""
		self.cmdline = cmdline
		self.working_directory = working_dir
		self.environment = environment
		self.error_string = ""
		self.return_code = 0

	def preproc_step(self):
		"""Re-implement this function in a derived class to have this function called
		directly before run() within the SimRunner parallel execution framework.
		"""
		pass

	def postproc_step(self):
		"""Re-implement this function in a derived class to have this function called
		directly after a successful run() within the SimRunner parallel execution framework.
		"""
		pass

	def run(self):
		"""Creates a new process to run the simulation.
		Returns when the simulation is complete.

		Function takes no arguments and returns the return code, which is also stored
		in the member variable return_code.

		**Return codes and error string**
		When a return code less than 0 is returned, the error_string member variable may hold
		additional information.

		When the working directory is invalid, return code is -2 returned.
		When the processes could not be spawned, possibly because of an invalid executable path, the
		code is -1 returned.
		"""

		if not self.working_directory == None:
			if not os.path.exists(self.working_directory):
				self.error_string = "Invalid working directory '{}'".format(self.working_directory)
				self.return_code = -2
				return self.return_code
		try:
			if platform.system() == "Windows":
				solver_process = subprocess.Popen(self.cmdline, creationflags=subprocess.CREATE_NEW_CONSOLE,
								                  cwd=self.working_directory, env=self.environment)
			else:
				solver_process = subprocess.Popen(self.cmdline,
								                  cwd=self.working_directory, env=self.environment)
			# run simulation, a successful simulation should return 0
			self.return_code = solver_process.wait()
		except Exception as ex:
			self.error_string = ex.message
			self.return_code = -1

		return self.return_code


class SimRunner:
	"""
	This class encapsulates a thread queue, a runner function and provides a convenient
	interface for running jobs in parallel.

	**Example:**

	.. code-block:: python

	   >>> from simulate import *

	   >>> # create job
	   >>> job1 = SimJob(['~/bin/DelphinSolver','-x','--verbosity-level=2','~/cases/D6/test/wall1.d6p'])

	   >>> # create another job
	   >>> job2 = SimJob(['~/bin/DelphinSolver','-x','--verbosity-level=2','~/cases/D6/test/wall2.d6p'])

	   >>> # put jobs in a list
	   >>> jobs = [job1, job2]

	   >>> # create parallel job runner
	   >>> runner = SimRunner(2) # 2 parallel threads

	   >>> # run jobs
	   >>> runner.run(jobs)

	   >>> # inspect return codes
	   >>> for j in runner.processed_jobs:
	   >>>     print "{} : {}".format(j.cmdline,j.return_code)
	"""

	def __init__(self, number_of_threads):
		"""
		Constructor.

		**Arguments**
		- *number_of_threads* - number of threads to use concurrently.
		"""
		# create a queue for task scheduling
		self.job_queue = Queue.Queue()
		self.processed_jobs = []
		self.number_of_threads = number_of_threads

	def thread_function(self):
		"""
		This function runs the simulation jobs.
		Do not call this function directly.
		"""

		# run as long as we have jobs in the Queue
		while 1:
			# get a new task from the Queue
			job = self.job_queue.get()
			# run preprocessing step
			job.preproc_step()
			# execute job
			ret_code = job.run()
			# if successful, also execute post-processing step
			if ret_code == 0:
				job.postproc_step()
			# store job and its result code to the list of processed jobs
			self.processed_jobs.append(job)
			# tell the Queue that the task was successfully done
			self.job_queue.task_done()

	def run(self, jobs):
		"""
		Function for adding jobs to the thread queue.
		jobs is a list/iteratable collection of SimJob objects.
		The function returns once all jobs have been processed.
		The exit codes of the jobs can be retrieved from the jobs.
		"""
		# now create as many threads as we need
		for i in range(self.number_of_threads):
			t = threading.Thread(target=SimRunner.thread_function,args=[self])
			t.daemon = True
			t.start()
		# now add simulation jobs to the queue
		for job in jobs:
			self.job_queue.put(job)
		# wait until all tasks are done
		self.job_queue.join()
		return
